"""Integration for building changes with GitLab CI.

Version Added:
    5.0
"""

from __future__ import annotations

import json
import logging
import re
from typing import Optional, TYPE_CHECKING

import gitlab
import pydantic
from django.utils.translation import gettext_lazy as _
from djblets.util.decorators import cached_property

from rbintegrations.baseci.errors import CIBuildError
from rbintegrations.baseci.integration import BaseCIIntegration, BuildPrepData
from rbintegrations.gitlabci.forms import (
    GitLabCIIntegrationConfigForm,
    TokenChoices,
)

if TYPE_CHECKING:
    from collections.abc import Mapping

    from djblets.util.typing import JSONDict, JSONValue
    from reviewboard.integrations.models import IntegrationConfig
    from reviewboard.reviews.models import ReviewRequest, StatusUpdate
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


class GitLabCIConfiguration(pydantic.BaseModel):
    """Configuration settings for the GitLab CI integration.

    These correspond to the fields in
    :py:class:`rbintegrations.gitlabci.forms.GitLabCIIntegrationConfigForm`.

    Version Added:
        5.0
    """

    #: The URL for the GitLab server.
    gitlab_endpoint: str

    #: JSON data for any inputs to pass to the pipeline.
    gitlab_inputs: str

    #: The GitLab project name or ID.
    gitlab_name: str

    #: The branch, tag, or other Git refname to use.
    gitlab_ref: str

    #: Whether to report build state in a review.
    gitlab_report_job_state: bool

    #: The type of token to use for invoking the pipeline.
    gitlab_token_type: TokenChoices

    #: The token content.
    gitlab_token: str

    #: JSON data for additional variables to pass to the pipeline.
    gitlab_vars: str

    #: The secret token to use for validating WebHook responses.
    gitlab_webhook_secret_token: Optional[str] = None


class GitLabCIIntegration(BaseCIIntegration):
    """Integration for building changes with GitLab CI.

    Version Added:
        5.0
    """

    name = 'GitLab CI'
    description = _('Builds diffs posted to Review Board using GitLab CI.')
    config_form_cls = GitLabCIIntegrationConfigForm

    status_update_service_id = 'gitlab-ci'

    bot_username = 'gitlab-ci'
    bot_user_first_name = 'GitLab'
    bot_user_last_name = 'CI'

    @cached_property
    def icon_static_urls(self) -> Mapping[str, str]:
        """The icons used for the integration.

        Type:
            dict
        """
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance
        assert extension is not None

        return {
            '1x': extension.get_static_url('images/gitlabci/icon.png'),
            '2x': extension.get_static_url('images/gitlabci/icon@2x.png'),
        }

    def start_build(
        self,
        *,
        prep_data: BuildPrepData,
        config: IntegrationConfig,
        status_update: StatusUpdate,
    ) -> None:
        """Start a new build.

        This will trigger a gitlab-ci build for the given review request and
        configuration.

        Args:
            prep_data (rbintegrations.baseci.integration.BuildPrepData):
                The builds preparation data.

            config (reviewboard.integrations.models.IntegrationConfig):
                The configuration for the integration triggering this build.

            status_update (reviewboard.reviews.models.StatusUpdate):
                The status update indicating the status of this build.

        Raises:
            rbintegrations.baseci.errors.CIBuildError:
                There was an error invoking the build.
        """
        diffset = prep_data.diffset
        local_site = prep_data.local_site
        repository = prep_data.repository
        review_request = prep_data.review_request

        try:
            gitlab_config = GitLabCIConfiguration.model_validate(
                config.settings)
        except pydantic.ValidationError as e:
            logger.error('Error validating GitLab CI configuration for '
                         'integration config pk=%s: %s',
                         config.pk, str(e))
            return

        name = self._replace_variables(
            name=gitlab_config.gitlab_name,
            repository=repository,
            review_request=review_request)

        logger.debug('Triggering GitLab CI build for review request '
                     '%s (diffset revision %d) on %s',
                     review_request.get_absolute_url(),
                     diffset.revision,
                     name)

        try:
            gitlab_vars = self._replace_variables_json(
                encoded=gitlab_config.gitlab_vars,
                repository=repository,
                review_request=review_request)
        except json.JSONDecodeError as e:
            logger.error('Unable to decode gitlab_vars in GitLab CI '
                         'integration configuration pk=%s: %s',
                         config.pk, str(e))
            return

        api_token = self.get_or_create_api_token(user=prep_data.user,
                                                 local_site=local_site)

        pipeline_name = (
            f'Review Request #{review_request.display_id}: '
            f'{review_request.summary}'
        )

        variables = {
            'REVIEWBOARD_API_TOKEN': api_token.token,
            'REVIEWBOARD_DIFF_REVISION': diffset.revision,
            'REVIEWBOARD_GITLAB_INTEGRATION_CONFIG_ID': config.pk,
            'REVIEWBOARD_PIPELINE_NAME': pipeline_name,
            'REVIEWBOARD_REVIEW_REQUEST': review_request.display_id,
            'REVIEWBOARD_SERVER': prep_data.server_url,
            'REVIEWBOARD_STATUS_UPDATE_ID': status_update.pk,
            **gitlab_vars,
        }

        try:
            inputs = self._replace_variables_json(
                encoded=gitlab_config.gitlab_inputs,
                repository=repository,
                review_request=review_request)
        except json.JSONDecodeError as e:
            logger.error('Unable to decode gitlab_inputs in GitLab CI '
                         'integration configuration pk=%s: %s',
                         config.pk, str(e))
            return

        target_ref = self._replace_variables(
            name=gitlab_config.gitlab_ref,
            repository=repository,
            review_request=review_request)

        try:
            if gitlab_config.gitlab_token_type == TokenChoices.TRIGGER_TOKEN:
                gl = gitlab.Gitlab(url=gitlab_config.gitlab_endpoint)
                project = gl.projects.get(name, lazy=True)

                pipeline = project.trigger_pipeline(
                    target_ref,
                    gitlab_config.gitlab_token,
                    inputs=inputs,
                    variables=variables,
                )
            else:
                gl = gitlab.Gitlab(url=gitlab_config.gitlab_endpoint,
                                   private_token=gitlab_config.gitlab_token)
                gl.auth()
                project = gl.projects.get(name)

                data: JSONDict = {
                    'inputs': inputs,
                    'ref': target_ref,
                    'variables': [
                        {
                            'key': key,
                            'value': value,
                        }
                        for key, value in variables.items()
                    ],
                }

                pipeline = project.pipelines.create(data)
        except Exception as e:
            raise CIBuildError(str(e))

        self.update_status(status_update,
                           url=pipeline.web_url,
                           url_text='View Pipeline')

    def _replace_variables_json(
        self,
        *,
        encoded: str | None,
        repository: Repository,
        review_request: ReviewRequest,
    ) -> JSONDict:
        """Replace variables in a JSON structure.

        Args:
            encoded (str):
                The encoded JSON string.

            repository (reviewboard.scmtools.models.Repository):
                The repository.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request.

        Returns:
            djblets.util.typing.JSONDict:
            The decoded JSON structure, with variables replaced in values.

        Raises:
            json.JSONDecodeError:
                The value could not be parsed as JSON.
        """
        if not encoded:
            return {}

        def _replace_in_value(
            value: JSONValue,
        ) -> JSONValue:
            if isinstance(value, str):
                return self._replace_variables(
                    name=value,
                    repository=repository,
                    review_request=review_request)
            elif isinstance(value, list):
                return [
                    self._replace_variables(
                        name=item,
                        repository=repository,
                        review_request=review_request)
                    for item in value
                ]
            else:
                return value

        return {
            key: _replace_in_value(value)
            for key, value in json.loads(encoded).items()
        }

    def _replace_variables(
        self,
        *,
        name: str,
        repository: Repository,
        review_request: ReviewRequest,
    ) -> str:
        """Replace variables in a configured name.

        This will replace the following variables:

        ``{branch}``:
            The branch name.

        ``{repository_name}``:
            The configured repository name.

        Args:
            name (str):
                The template string for the name.

            repository (reviewboard.scmtools.models.Repository):
                The repository for the change being built.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request for the change being built.

        Returns:
            str:
            The resulting name.
        """
        var_map = {
            'repository_name': repository.name,
            'branch': review_request.branch,
        }

        keys = '|'.join(re.escape(key) for key in var_map)

        return re.sub(
            fr'\{{({keys})\}}',
            lambda m: var_map[m.group(1)],
            name)



"""Unit tests for the Travis CI integration."""

import json
from urllib.error import HTTPError

from django.urls import reverse
from djblets.conditions import ConditionSet, Condition
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.hostingsvcs.service import (HostingServiceHTTPRequest,
                                             HostingServiceHTTPResponse)
from reviewboard.reviews.conditions import ReviewRequestRepositoriesChoice
from reviewboard.reviews.models import StatusUpdate
from reviewboard.reviews.signals import status_update_request_run

from rbintegrations.travisci.api import TravisAPI
from rbintegrations.travisci.forms import (GitHubRepositoriesChoice,
                                           TravisCIIntegrationConfigForm)
from rbintegrations.travisci.integration import TravisCIIntegration
from rbintegrations.travisci.views import TravisCIWebHookView
from rbintegrations.testing.testcases import IntegrationTestCase


class BaseTravisCITestCase(IntegrationTestCase):
    """Base class for Travis CI tests."""

    integration_cls = TravisCIIntegration
    fixtures = ['test_scmtools', 'test_users']

    def _create_repository(self, github=True, repository_plan='public-org'):
        """Create and return a repository for testing.

        Args:
            github (bool, optional):
                Whether the repository should use the GitHub hosting service.

            repository_plan (unicode, optional):
                The type of GitHub repository plan.

        Returns:
            reviewboard.scmtools.models.Repository:
            A repository for use in unit tests.
        """
        if github:
            account = HostingServiceAccount(service_name='github',
                                            username='myuser')

            def _http_get_user(_self, url, *args, **kwargs):
                self.assertEqual(url, 'https://api.github.com/user')

                payload = b'{}'
                headers = {
                    str('X-OAuth-Scopes'): str('admin:repo_hook, repo, user'),
                }

                return HostingServiceHTTPResponse(
                    request=HostingServiceHTTPRequest(url=url),
                    url=url,
                    data=payload,
                    headers=headers,
                    status_code=200)

            service = account.service
            self.spy_on(service.client.http_get,
                        call_fake=_http_get_user)

            service.authorize('myuser', 'mypass', None)
            self.assertTrue(account.is_authorized)

            service.client.http_get.unspy()

            repository = self.create_repository()
            repository.hosting_account = account
            repository.extra_data['repository_plan'] = repository_plan

            if repository_plan == 'public':
                repository.extra_data['github_public_repo_name'] = \
                    'mypublicrepo'
            elif repository_plan == 'public-org':
                repository.extra_data['github_public_org_name'] = 'mypublicorg'
                repository.extra_data['github_public_org_repo_name'] = \
                    'mypublicorgrepo'
            elif repository_plan == 'private':
                repository.extra_data['github_private_repo_name'] = \
                    'myprivaterepo'
            elif repository_plan == 'private-org':
                repository.extra_data['github_private_org_name'] = \
                    'myprivateorg'
                repository.extra_data['github_private_org_repo_name'] = \
                    'myprivateorgrepo'

            repository.save()
            return repository
        else:
            return self.create_repository()

    def _create_config(self, enterprise=False, with_local_site=False,
                       run_manually=False):
        """Create an integration config.

        Args:
            enterprise (bool, optional):
                Whether to use an enterprise endpoint or the default
                open-source endpoint.

            with_local_site (bool, optional):
                Whether to limit the config to a local site.
        """
        choice = ReviewRequestRepositoriesChoice()

        condition_set = ConditionSet(conditions=[
            Condition(choice=choice,
                      operator=choice.get_operator('any'))
        ])

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        config = self.integration.create_config(name='Config 1',
                                                enabled=True,
                                                local_site=local_site)
        config.set('conditions', condition_set.serialize())
        config.set('travis_yml', 'script:\n    python ./tests/runtests.py')
        config.set('branch_name', 'review-requests')
        config.set('run_manually', run_manually)

        if enterprise:
            config.set('travis_endpoint', TravisAPI.ENTERPRISE_ENDPOINT)
            config.set('travis_custom_endpoint', 'https://travis.example.com/')
        else:
            config.set('travis_endpoint', TravisAPI.OPEN_SOURCE_ENDPOINT)

        config.save()

        return config

    def _spy_on_make_request(self):
        """Wrapper function for spying on the urlopen function.

        Returns:
            dict:
            Faked response from TravisCI.
        """
        data = {}

        def _make_request(api, url, body=None, method='GET', headers={},
                          content_type=None):
            # We can't actually do any assertions in here, because they'll get
            # swallowed by SignalHook's sandboxing. We therefore record the
            # data we need and assert later.
            data['url'] = url
            data['request'] = json.loads(body)['request']
            return '{}'

        self.spy_on(TravisAPI._make_request, owner=TravisAPI,
                    call_fake=_make_request)

        return data


class GitHubRepositoriesChoiceTests(BaseTravisCITestCase):
    """Unit tests for GitHubRepositoriesChoice.

    Version Added:
        4.0
    """

    def test_get_queryset(self) -> None:
        """Testing GitHubRepositoriesChoice.get_queryset"""
        local_site = self.create_local_site()

        github_account1 = HostingServiceAccount.objects.create(
            service_name='github',
            username='myuser1')
        github_account2 = HostingServiceAccount.objects.create(
            service_name='github',
            username='myuser2')
        github_account3 = HostingServiceAccount.objects.create(
            service_name='github',
            username='myuser2',
            local_site=local_site)

        # These should match.
        repo1 = self.create_repository(name='repo1',
                                       hosting_account=github_account1)
        repo2 = self.create_repository(name='repo2',
                                       hosting_account=github_account2)

        # These should not match.
        self.create_repository(name='repo3',
                               tool_name='Git')
        self.create_repository(name='repo4',
                               hosting_account=github_account3,
                               local_site=local_site)

        choice = GitHubRepositoriesChoice()
        choice.extra_state.update({
            'local_site': None,
            'matching': True,
        })

        self.assertQuerySetEqual(
            choice.get_queryset().order_by('pk'),
            [repo1, repo2])

    def test_get_queryset_with_local_site(self) -> None:
        """Testing GitHubRepositoriesChoice.get_queryset with Local Site"""
        good_local_site = self.create_local_site(name='good-site')
        bad_local_site = self.create_local_site(name='bad-site')

        github_account1 = HostingServiceAccount.objects.create(
            service_name='github',
            username='myuser1',
            local_site=good_local_site)
        github_account2 = HostingServiceAccount.objects.create(
            service_name='github',
            username='myuser2',
            local_site=good_local_site)
        github_account3 = HostingServiceAccount.objects.create(
            service_name='github',
            username='myuser2',
            local_site=bad_local_site)

        # These should match.
        repo1 = self.create_repository(name='repo1',
                                       hosting_account=github_account1,
                                       local_site=good_local_site)
        repo2 = self.create_repository(name='repo2',
                                       hosting_account=github_account2,
                                       local_site=good_local_site)

        # These should not match.
        self.create_repository(name='repo3',
                               tool_name='Git')
        self.create_repository(name='repo4',
                               tool_name='Git',
                               local_site=good_local_site)
        self.create_repository(name='repo5',
                               hosting_account=github_account3,
                               local_site=bad_local_site)

        choice = GitHubRepositoriesChoice()
        choice.extra_state.update({
            'local_site': good_local_site,
            'matching': True,
        })

        self.assertQuerySetEqual(
            choice.get_queryset().order_by('pk'),
            [repo1, repo2])


class TravisCIIntegrationTests(BaseTravisCITestCase):
    """Tests for Travis CI."""

    def test_build_new_review_request(self):
        """Testing TravisCIIntegration builds a new review request"""
        repository = self._create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        config = self._create_config()
        self.integration.enable_integration()

        data = self._spy_on_make_request()

        review_request.publish(review_request.submitter)

        self.assertTrue(TravisAPI._make_request.called)

        self.assertEqual(
            data['url'],
            'https://api.travis-ci.org/repo/mypublicorg%2Fmypublicorgrepo/'
            'requests')

        self.assertEqual(
            data['request']['config']['env']['global'],
            [
                'REVIEWBOARD_STATUS_UPDATE_ID=1',
                'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d' % config.pk,
            ])

        self.assertEqual(data['request']['message'],
                         'Test Summary\n\nTest Description')
        self.assertTrue('git fetch --unshallow origin || true'
                        in data['request']['config']['before_install'])
        self.assertTrue('git checkout %s' % diffset.base_commit_id
                        in data['request']['config']['before_install'])
        self.assertEqual(data['request']['branch'], 'review-requests')

    def test_build_new_review_request_on_enterprise_travis(self):
        """Testing TravisCIIntegration builds a new review request with Travis
        CI Enterprise
        """
        repository = self._create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        config = self._create_config(enterprise=True)
        self.integration.enable_integration()

        data = self._spy_on_make_request()

        review_request.publish(review_request.submitter)

        self.assertTrue(TravisAPI._make_request.called)

        self.assertEqual(
            data['url'],
            'https://travis.example.com/api/repo/'
            'mypublicorg%2Fmypublicorgrepo/requests')

        self.assertEqual(
            data['request']['config']['env']['global'],
            [
                'REVIEWBOARD_STATUS_UPDATE_ID=1',
                'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d' % config.pk,
            ])
        self.assertEqual(data['request']['message'],
                         'Test Summary\n\nTest Description')
        self.assertTrue('git checkout %s' % diffset.base_commit_id
                        in data['request']['config']['before_install'])
        self.assertEqual(data['request']['branch'], 'review-requests')

    def test_build_new_review_request_with_parent_diff(self):
        """Testing TravisCIIntegration build script with a parent diff"""
        repository = self._create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        filediff = self.create_filediff(diffset)
        filediff.parent_diff = (
            b'--- README\trevision 123\n'
            b'+++ README\trevision 123\n'
            b'@@ -1 +1 @@\n'
            b'-Hello, world!\n'
            b'+Hello, everybody!\n'
        )
        filediff.save()

        config = self._create_config(enterprise=True)
        self.integration.enable_integration()

        data = self._spy_on_make_request()

        review_request.publish(review_request.submitter)

        self.assertTrue(TravisAPI._make_request.called)

        self.assertEqual(
            data['url'],
            'https://travis.example.com/api/repo/'
            'mypublicorg%2Fmypublicorgrepo/requests')

        self.assertEqual(
            data['request']['config']['env']['global'],
            [
                'REVIEWBOARD_STATUS_UPDATE_ID=1',
                'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d' % config.pk,
            ])
        self.assertEqual(data['request']['message'],
                         'Test Summary\n\nTest Description')
        self.assertTrue('git checkout %s' % diffset.base_commit_id
                        in data['request']['config']['before_install'])
        self.assertEqual(data['request']['branch'], 'review-requests')

        patch_count = len(
            [cmd for cmd in data['request']['config']['before_install']
             if 'patch -p1' in cmd])

        self.assertEqual(patch_count, 2)

    def test_build_new_review_request_with_git_depth(self):
        """Testing TravisCIIntegration builds with git: depth: False"""
        repository = self._create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        config = self._create_config()
        config.set('travis_yml',
                   'git:\n'
                   '    depth:\n'
                   '        False\n'
                   '\n'
                   'script:\n'
                   '    python ./tests/runtests.py')
        config.save()
        self.integration.enable_integration()

        data = self._spy_on_make_request()

        review_request.publish(review_request.submitter)

        self.assertTrue(TravisAPI._make_request.called)

        self.assertEqual(
            data['url'],
            'https://api.travis-ci.org/repo/mypublicorg%2Fmypublicorgrepo/'
            'requests')

        self.assertEqual(
            data['request']['config']['env']['global'],
            [
                'REVIEWBOARD_STATUS_UPDATE_ID=1',
                'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d' % config.pk,
            ])
        self.assertEqual(data['request']['message'],
                         'Test Summary\n\nTest Description')
        self.assertFalse('git fetch --unshallow origin || true'
                         in data['request']['config']['before_install'])
        self.assertEqual(data['request']['branch'], 'review-requests')

    def test_non_github_review_request(self):
        """Testing TravisCIIntegration skipping a non-GitHub review request"""
        repository = self._create_repository(github=False)
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config()
        self.integration.enable_integration()

        self.spy_on(TravisAPI._make_request, owner=TravisAPI,
                    call_original=False)

        review_request.publish(review_request.submitter)

        self.assertFalse(TravisAPI._make_request.called)

    def test_travisci_config_form_valid(self):
        """Testing TravisCIIntegrationConfigForm validation success"""
        form = TravisCIIntegrationConfigForm(
            integration=self.integration,
            request=None,
            data={
                'conditions_last_id': 0,
                'conditions_mode': 'always',
                'name': 'test',
                'travis_endpoint': TravisAPI.OPEN_SOURCE_ENDPOINT,
                'travis_ci_token': '123456',
                'travis_yml': 'script:\n    - python ./tests/runtests.py',
            })

        self.spy_on(TravisAPI.get_user, owner=TravisAPI, call_original=False)
        self.spy_on(TravisAPI.lint,
                    owner=TravisAPI,
                    call_fake=lambda x, travis_yml: {'warnings': []})

        self.assertTrue(form.is_valid())

    def test_travisci_config_form_lint_failure(self):
        """Testing TravisCIIntegrationConfigForm validation lint failure"""
        self.spy_on(TravisAPI.get_user, owner=TravisAPI, call_original=False)
        self.spy_on(
            TravisAPI.lint,
            owner=TravisAPI,
            call_fake=lambda x, travis_yml: {
                'warnings': [{
                    'key': 'script',
                    'message': 'An error',
                }],
            })

        form = TravisCIIntegrationConfigForm(
            integration=self.integration,
            request=None,
            data={
                'conditions_last_id': 0,
                'conditions_mode': 'always',
                'name': 'test',
                'travis_endpoint': TravisAPI.OPEN_SOURCE_ENDPOINT,
                'travis_ci_token': '123456',
                'travis_yml': 'script:\n    - python ./tests/runtests.py',
            })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['travis_yml'],
                         ['In script section: An error'])

    def test_travisci_config_form_auth_failure(self):
        """Testing TravisCIIntegrationConfigForm config validation"""
        def _raise_403(obj):
            raise HTTPError('', 403, 'Authentication failed', None, None)

        self.spy_on(TravisAPI.get_user, owner=TravisAPI, call_fake=_raise_403)

        form = TravisCIIntegrationConfigForm(
            integration=self.integration,
            request=None,
            data={
                'conditions_last_id': 0,
                'conditions_mode': 'always',
                'name': 'test',
                'travis_endpoint': TravisAPI.OPEN_SOURCE_ENDPOINT,
                'travis_ci_token': '123456',
                'travis_yml': 'script:\n    - python ./tests/runtests.py',
            })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['travis_ci_token'],
                         ['Unable to authenticate with this API token.'])

    def test_manual_run_no_build_on_publish(self):
        """Testing lack of TravisCIIntegration build when a new review
        request is made with the run manually configuration
        """
        repository = self._create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config(run_manually=True)
        self.integration.enable_integration()

        data = self._spy_on_make_request()

        review_request.publish(review_request.submitter)

        self.assertFalse(TravisAPI._make_request.called)

        self.assertEqual(data, {})

    def test_build_manual_run(self):
        """Testing TravisCIIntegration build via a manual trigger"""
        repository = self._create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        config = self._create_config()
        self.integration.enable_integration()

        status_update = \
            self.create_status_update(service_id='travis-ci',
                                      review_request=review_request)

        data = self._spy_on_make_request()

        status_update_request_run.send(sender=self.__class__,
                                       status_update=status_update)

        self.assertTrue(TravisAPI._make_request.called)

        self.assertEqual(
            data['url'],
            'https://api.travis-ci.org/repo/'
            'mypublicorg%2Fmypublicorgrepo/requests')

        self.assertEqual(
            data['request']['config']['env']['global'],
            [
                'REVIEWBOARD_STATUS_UPDATE_ID=1',
                'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d' % config.pk,
            ])

        self.assertEqual(data['request']['message'],
                         'Test Summary\n\nTest Description')
        self.assertTrue('git fetch --unshallow origin || true'
                        in data['request']['config']['before_install'])
        self.assertTrue('git checkout %s' % diffset.base_commit_id
                        in data['request']['config']['before_install'])
        self.assertEqual(data['request']['branch'], 'review-requests')

    def test_build_new_review_request_with_public_github_repository(self):
        """Testing CircleCIIntegration builds for a new review request with
        a public GitHub repository
        """
        repository = self._create_repository(repository_plan='public')
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config()
        self.integration.enable_integration()

        data = self._spy_on_make_request()

        review_request.publish(review_request.submitter)

        self.assertTrue(TravisAPI._make_request.called)

        self.assertEqual(
            data['url'],
            'https://api.travis-ci.org/repo/myuser%2Fmypublicrepo/requests')

    def test_build_new_review_request_with_private_github_repository(self):
        """Testing CircleCIIntegration builds for a new review request with
        a public GitHub repository
        """
        repository = self._create_repository(repository_plan='private')
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config()
        self.integration.enable_integration()

        data = self._spy_on_make_request()

        review_request.publish(review_request.submitter)

        self.assertTrue(TravisAPI._make_request.called)

        self.assertEqual(
            data['url'],
            'https://api.travis-ci.org/repo/myuser%2Fmyprivaterepo/requests')

    def test_build_new_review_request_with_private_org_github_repository(self):
        """Testing CircleCIIntegration builds for a new review request with
        a public GitHub repository
        """
        repository = self._create_repository(repository_plan='private-org')
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request)
        diffset.base_commit_id = '8fd69d70f07b57c21ad8733c1c04ae604d21493f'
        diffset.save()

        self._create_config()
        self.integration.enable_integration()

        data = self._spy_on_make_request()

        review_request.publish(review_request.submitter)

        self.assertTrue(TravisAPI._make_request.called)

        self.assertEqual(
            data['url'],
            'https://api.travis-ci.org/repo/'
            'myprivateorg%2Fmyprivateorgrepo/requests')


class TravisCIWebHookTests(BaseTravisCITestCase):
    """Tests for the Travis CI webhook handler."""

    def setUp(self):
        super(TravisCIWebHookTests, self).setUp()

        self.repository = self._create_repository()
        self.review_request = self.create_review_request(
            repository=self.repository)

        self.status_update = self.create_status_update(self.review_request)

        self.config = self._create_config()
        self.integration.enable_integration()

        self.webhook_url = reverse('travis-ci-webhook')

    def test_webhook_no_env(self):
        """Testing TravisCIWebHookView with missing env"""
        payload = json.dumps({})
        rsp = self.client.post(self.webhook_url, {'payload': payload})

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(rsp.content, b'Got event without an env in config.')

    def test_webhook_missing_ids(self):
        """Testing TravisCIWebHookView with missing object IDs"""
        payload = json.dumps({
            'matrix': [
                {
                    'config': {
                        'env': [
                            'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d'
                            % self.config.pk,
                        ],
                    },
                },
            ],
        })
        rsp = self.client.post(self.webhook_url, {'payload': payload})

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(
            rsp.content,
            b'Unable to find REVIEWBOARD_STATUS_UPDATE_ID in payload.')

        payload = json.dumps({
            'matrix': [
                {
                    'config': {
                        'env': [
                            'REVIEWBOARD_STATUS_UPDATE_ID=%d'
                            % self.status_update.pk,
                        ],
                    },
                },
            ],
        })
        rsp = self.client.post(self.webhook_url, {'payload': payload})

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(
            rsp.content,
            b'Unable to find REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID in '
            b'payload.')

    def test_webhook_bad_integration_config(self):
        """Testing TravisCIWebHookView with incorrect integration config ID"""
        payload = json.dumps({
            'matrix': [
                {
                    'config': {
                        'env': [
                            'REVIEWBOARD_STATUS_UPDATE_ID=%d'
                            % self.status_update.pk,
                            'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d'
                            % (self.config.pk + 1),
                        ],
                    },
                },
            ],
        })
        rsp = self.client.post(self.webhook_url, {'payload': payload})

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(
            rsp.content,
            b'Unable to find matching integration config ID %d.'
            % (self.config.pk + 1))

    def test_webhook_bad_signature(self):
        """Testing TravisCIWebHookView with bad HTTP_SIGNATURE"""
        payload = json.dumps({
            'matrix': [
                {
                    'config': {
                        'env': [
                            'REVIEWBOARD_STATUS_UPDATE_ID=%d'
                            % self.status_update.pk,
                            'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d'
                            % self.config.pk,
                        ],
                    },
                },
            ],
        })
        rsp = self.client.post(self.webhook_url, {'payload': payload})

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(
            rsp.content,
            b'Invalid Travis CI webhook signature for status update %d.'
            % self.status_update.pk)

    def test_webhook_bad_status_update(self):
        """Testing TravisCIWebHookView with incorrect status update ID"""
        payload = json.dumps({
            'matrix': [
                {
                    'config': {
                        'env': [
                            'REVIEWBOARD_STATUS_UPDATE_ID=%d'
                            % (self.status_update.pk + 1),
                            'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d'
                            % self.config.pk,
                        ],
                    },
                },
            ],
        })
        self.spy_on(TravisCIWebHookView._validate_signature,
                    owner=TravisCIWebHookView,
                    call_fake=lambda self, request, integration_config: True)

        rsp = self.client.post(self.webhook_url, {'payload': payload})

        self.assertEqual(rsp.status_code, 400)
        self.assertEqual(
            rsp.content,
            b'Unable to find matching status update ID %d.'
            % (self.status_update.pk + 1))

    def test_webhook_build_pending(self):
        """Testing TravisCIWebHookView build pending"""
        payload = json.dumps({
            'matrix': [
                {
                    'config': {
                        'env': [
                            'REVIEWBOARD_STATUS_UPDATE_ID=%d'
                            % self.status_update.pk,
                            'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d'
                            % self.config.pk,
                        ],
                    },
                },
            ],
            'build_url': 'https://example.com/build',
            'state': 'started',
        })
        self.spy_on(TravisCIWebHookView._validate_signature,
                    owner=TravisCIWebHookView,
                    call_fake=lambda self, request, integration_config: True)

        rsp = self.client.post(self.webhook_url, {'payload': payload})

        self.assertEqual(rsp.status_code, 200)

        self.status_update = StatusUpdate.objects.get(pk=self.status_update.pk)
        self.assertEqual(self.status_update.url, 'https://example.com/build')
        self.assertEqual(self.status_update.state,
                         StatusUpdate.PENDING)

    def test_webhook_build_success(self):
        """Testing TravisCIWebHookView build success"""
        payload = json.dumps({
            'matrix': [
                {
                    'config': {
                        'env': [
                            'REVIEWBOARD_STATUS_UPDATE_ID=%d'
                            % self.status_update.pk,
                            'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d'
                            % self.config.pk,
                        ],
                    },
                },
            ],
            'build_url': 'https://example.com/build',
            'state': 'passed',
        })
        self.spy_on(TravisCIWebHookView._validate_signature,
                    owner=TravisCIWebHookView,
                    call_fake=lambda self, request, integration_config: True)

        rsp = self.client.post(self.webhook_url, {'payload': payload})

        self.assertEqual(rsp.status_code, 200)

        self.status_update = StatusUpdate.objects.get(pk=self.status_update.pk)
        self.assertEqual(self.status_update.url, 'https://example.com/build')
        self.assertEqual(self.status_update.state,
                         StatusUpdate.DONE_SUCCESS)

    def test_webhook_build_error(self):
        """Testing TravisCIWebHookView build error"""
        payload = json.dumps({
            'matrix': [
                {
                    'config': {
                        'env': [
                            'REVIEWBOARD_STATUS_UPDATE_ID=%d'
                            % self.status_update.pk,
                            'REVIEWBOARD_TRAVIS_INTEGRATION_CONFIG_ID=%d'
                            % self.config.pk,
                        ],
                    },
                },
            ],
            'build_url': 'https://example.com/build',
            'state': 'failed',
        })
        self.spy_on(TravisCIWebHookView._validate_signature,
                    owner=TravisCIWebHookView,
                    call_fake=lambda self, request, integration_config: True)

        rsp = self.client.post(self.webhook_url, {'payload': payload})

        self.assertEqual(rsp.status_code, 200)

        self.status_update = StatusUpdate.objects.get(pk=self.status_update.pk)
        self.assertEqual(self.status_update.url, 'https://example.com/build')
        self.assertEqual(self.status_update.state,
                         StatusUpdate.DONE_FAILURE)



"""Errors for builds.

Version Added:
    3.1
"""

from typing import Optional


class CIBuildError(Exception):
    """An error invoking a build over CI.

    Version Added:
        3.1
    """

    ######################
    # Instance variables #
    ######################

    #: An optional URL containing information on the build.
    #:
    #: Type:
    #:     str
    url: Optional[str]

    #: The text for an optional URL containing information on the build.
    #:
    #: Type:
    #:     str
    url_text: Optional[str]

    def __init__(
        self,
        message: Optional[str] = None,
        *,
        url: Optional[str] = None,
        url_text: Optional[str] = None,
    ) -> None:
        """Initialize the error.

        Args:
            message (str, optional):
                An explicit error message for the error. This will be shown
                on the status update, and should follow standard conventions
                (start as lowercase, end with a period).

                If not provided, a default is used instead.

            url (str, optional):
                An optional URL containing information on the build.

            url_text (str, optional):
                The text for an optional URL containing information on the
                build.
        """
        super().__init__(message or 'error starting the build.')

        self.url = url
        self.url_text = url_text



"""Integration for Microsoft Teams.

Version Added:
    4.0
"""

from __future__ import annotations

import json
import logging
import re
from typing import ClassVar, MutableMapping, Optional, Sequence, TYPE_CHECKING
from urllib.request import Request, urlopen

from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from reviewboard.admin.server import build_server_url

from rbintegrations.basechat.integration import BaseChatIntegration
from rbintegrations.msteams.forms import MSTeamsIntegrationConfigForm
from rbintegrations.util.compat.logs import log_timed

if TYPE_CHECKING:
    from djblets.util.typing import JSONDict, JSONList
    from rbintegrations.basechat.integration import FieldsDict
    from reviewboard.site.models import LocalSite
    from reviewboard.reviews.models import ReviewRequest


logger = logging.getLogger(__name__)


class MSTeamsIntegration(BaseChatIntegration):
    """Integrates Review Board with MS Teams.

    This will handle updating MS Teams channels when review requests
    are posted, changed, or closed, and when there's new activity on the
    review request.

    Version Added:
        4.0
    """

    name = 'Microsoft Teams'
    description = _(
        'Notifies channels in Microsoft Teams when review requests are '
        'created, updated, and reviewed.'
    )

    default_settings = {
        'notify_username': 'Review Board',
        'webhook_url': '',
    }

    config_form_cls = MSTeamsIntegrationConfigForm

    use_emoji_shortcode = False

    assets_base_url: ClassVar[str] = (
        'https://static.reviewboard.org/integration-assets/msteams'
    )

    def notify(
        self,
        *,
        title: str,
        title_link: str,
        fallback_text: str,
        local_site: Optional[LocalSite],
        review_request: ReviewRequest,
        event_name: Optional[str] = None,
        fields: Sequence[FieldsDict] = [],
        pre_text: Optional[str] = None,
        body: Optional[str] = None,
        color: Optional[str] = None,
        thumb_url: Optional[str] = None,
        image_url: Optional[str] = None,
    ) -> None:
        """Send a WebHook notification to MS Teams.

        This will post the given message to any MS Teams channels
        configured to receive it.

        The message JSON payload follows the Adaptive Card v1.5 schema.

        Args:
            title (str):
                The title for the message.

            title_link (str):
                The link for the title of the message.

            fallback_text (str):
                The non-rich fallback text to display in the chat, for use in
                IRC and other services.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site for the review request or review emitting
                the message. Only integration configurations matching this
                Local Site will be processed.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request the notification is bound to.

            event_name (str, optional):
                The name of the event triggering this notification.

            fields (list of rbintegrations.basechat.integration.FieldsDict,
                    optional):
                The fields comprising the rich message to display in chat.

            pre_text (str, optional):
                Text to display before the rest of the message.

            body (str, optional):
                The body of the message.

            color (str, optional):
                An RGB hex value for the message.

            thumb_url (str, optional):
                URL of an image to show on the side of the message.

            image_url (str, optional):
                URL of an image to show in the message.
        """
        pre_text_card_columns: JSONList = [
            {
                'type': 'Column',
                'width': 'Auto',
                'verticalContentAlignment': 'Center',
                'items': [
                    {
                        'type': 'TextBlock',
                        'text': pre_text,
                        'size': 'Medium',
                        'style': 'Heading',
                    },
                ]
            }
        ]

        if thumb_url:
            pre_text_card_columns.append({
                'type': 'Column',
                'width': 'auto',
                'spacing': 'None',
                'items': [
                    {
                        'type': 'Image',
                        'url': thumb_url,
                        'altText': 'A trophy earned on the review request.',
                        'width': '32px',
                        'height': '32px',
                    }
                ],
            })

        pre_text_card: JSONDict = {
            'type': 'ColumnSet',
            'columns': pre_text_card_columns
        }

        activity_card: JSONDict = {
            'type': 'ColumnSet',
            'columns': [
                {
                    'type': 'Column',
                    'width': 'Auto',
                    'items': [
                        {
                            'type': 'Image',
                            'url': self.logo_url,
                            'alt': 'Review Board notification',
                            'width': '45px',
                            'height': '45px',
                        },
                    ],
                },
                {
                    'type': 'Column',
                    'items': [
                        {
                            'type': 'TextBlock',
                            'text': self.format_link(path=title_link,
                                                     text=title),
                            'weight': 'Bolder',
                            'wrap': True,
                        },
                        {
                            'type': 'TextBlock',
                            'text': body or '',
                            'isSubtle': True,
                            'spacing': 'Small',
                            'wrap': True,
                        },
                    ]
                },
            ]
        }

        main_body: list[JSONDict] = [
            pre_text_card,
            activity_card,
        ]

        if fields:
            main_body.append({
                'type': 'FactSet',
                'facts': [
                    {
                        'title': field['title'],
                        'value': field['value'],
                    }
                    for field in fields
                ],
            })

        if image_url:
            main_body.append({
                'type': 'Image',
                'url': image_url,
                'altText': 'An image file attachment from the review request.',
            })

        payload: JSONDict = {
            'type': 'message',
            'attachments': [
                {
                    'contentType': 'application/vnd.microsoft.card.adaptive',
                    'contentUrl': None,
                    'content':
                        {
                            '$schema': 'http://adaptivecards.io/schemas/'
                                       'adaptive-card.json',
                            'type': 'AdaptiveCard',
                            'version': '1.5',
                            'body': main_body,
                        },
                },
            ],
        }

        # Send a notification to any configured channels.
        for config in self.get_configs(local_site):
            if not config.match_conditions(form_cls=self.config_form_cls,
                                           review_request=review_request):
                continue

            webhook_url = config.get('webhook_url')

            with log_timed(f'Sending notification for event "{event_name}", '
                           f'review_request ID {review_request.pk}, '
                           f'WebHook URL {webhook_url}',
                           logger=logger) as log_timer:
                try:
                    if not webhook_url:
                        raise Exception('WebHook URL has not been configured.')

                    data = json.dumps(payload).encode('utf-8')
                    headers: MutableMapping[str, str] = {
                        'Content-Length': str(len(data)),
                        'Content-Type': 'application/json',
                    }
                    urlopen(Request(webhook_url, data, headers))
                except Exception as e:
                    logger.exception('[%s] Failed to send notification: %s',
                                     log_timer.trace_id, e)

    def format_link(
        self,
        *,
        path: str,
        text: str,
    ) -> str:
        """Format the given URL and text to be shown in a MS Teams message.

        This will combine together the parts of the URL (method, domain, path)
        and format it using MS Teams' URL syntax.

        Args:
            path (str):
                The path on the Review Board server.

            text (str):
                The text for the link.

        Returns:
            str:
            The link for use in MS Teams.
        """
        # We only care about escaping parentheses in the URL, since those are
        # the only things that can break the Markdown link.
        escape_map = {
            '(': '%28',
            ')': '%29',
        }

        path = re.sub(
            '[()]',
            lambda m: escape_map[m.group(0)],
            path)

        # We only care about escaping brackets in the text, since those
        # are the only things that can break the Markdown link.
        text = re.sub(r'([\[\]])', r'\\\1', text)

        return (
            f'[{text}]'
            f'({build_server_url(path)})'
        )

    # NOTE: It's technically type-unsafe to replace an attribute with a
    #       property, but this works for us practically.
    @cached_property
    def icon_static_urls(self) -> dict[str, str]:  # type: ignore
        """Return the icons used for the integration.

        Returns:
            dict:
            The icons for MS Teams.
        """
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        assert extension is not None

        return {
            '1x': extension.get_static_url('images/msteams/icon.png'),
            '2x': extension.get_static_url('images/msteams/icon@2x.png'),
        }



"""Integration for building changes on Jenkins CI."""

from __future__ import annotations

import logging
import re
from typing import Dict, TYPE_CHECKING
from urllib.error import HTTPError

from djblets.util.decorators import cached_property

from rbintegrations.baseci.errors import CIBuildError
from rbintegrations.baseci.integration import BaseCIIntegration, BuildPrepData
from rbintegrations.jenkinsci.api import JenkinsAPI
from rbintegrations.jenkinsci.forms import JenkinsCIIntegrationConfigForm

if TYPE_CHECKING:
    from reviewboard.integrations.models import IntegrationConfig
    from reviewboard.reviews.models import ReviewRequest, StatusUpdate
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


class JenkinsCIIntegration(BaseCIIntegration):
    """Integrates Review Board with Jenkins CI."""

    name = 'Jenkins CI'
    description = 'Builds diffs posted to Review Board using Jenkins CI.'
    config_form_cls = JenkinsCIIntegrationConfigForm

    status_update_service_id = 'jenkins-ci'

    bot_username = 'jenkins-ci'
    bot_user_first_name = 'Jenkins'
    bot_user_last_name = 'CI'

    @cached_property
    def icon_static_urls(self) -> Dict:
        """The icons used for the integration.

        Type:
            dict
        """
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        return {
            '1x': extension.get_static_url('images/jenkinsci/icon.png'),
            '2x': extension.get_static_url('images/jenkinsci/icon@2x.png'),
        }

    def prepare_builds(
        self,
        prep_data: BuildPrepData,
    ) -> bool:
        """Prepare for builds.

        This will set some initial information common to all Jenkins builds.

        Version Added:
            3.1

        Args:
            prep_data (BuildPrepData):
                The builds preparation data to modify.

        Returns:
            bool:
            ``True``, always.
        """
        review_request = prep_data.review_request

        prep_data.extra_state['patch_info'] = {
            'diff_revision': prep_data.diffset.revision,
            'review_branch': review_request.branch,
            'review_id': review_request.display_id,
            'reviewboard_server': prep_data.server_url,
        }

        return True

    def start_build(
        self,
        *,
        prep_data: BuildPrepData,
        config: IntegrationConfig,
        status_update: StatusUpdate,
    ) -> None:
        """Start a new build.

        This will trigger a Jenkins build for the given review request and
        configuration.

        Version Added:
            3.1

        Args:
            prep_data (BuildPrepData):
                The builds preparation data containing information for the
                build.

            config (reviewboard.integrations.models.IntegrationConfig):
                The configuration for the integration triggering this build.

            status_update (reviewboard.reviews.models.status_update.
                           StatusUpdate):
                The status update indicating the status of this build.

        Raises:
            rbintegrations.baseci.errors.CIBuildError:
                There was an error invoking the build.
        """
        patch_info = prep_data.extra_state['patch_info'].copy()
        patch_info['status_update_id'] = status_update.pk

        diffset = prep_data.diffset
        repository = prep_data.repository
        review_request = prep_data.review_request

        job_name = self._replace_job_variables(config.get('jenkins_job_name'),
                                               repository,
                                               review_request)

        # Time to kick off the build!
        logger.info('Triggering Jenkins CI build for review request '
                    '%s (diffset revision %d)',
                    review_request.get_absolute_url(),
                    diffset.revision)

        api = JenkinsAPI(endpoint=config.get('jenkins_endpoint'),
                         job_name=job_name,
                         username=config.get('jenkins_username'),
                         password=config.get('jenkins_password'))

        try:
            api.start_build(patch_info)
        except HTTPError as e:
            if e.code == 404:
                raise CIBuildError('failed, job does not exist.')
            else:
                raise CIBuildError('failed to communicate with Jenkins.')

    def _replace_job_variables(
        self,
        job_name: str,
        repository: Repository,
        review_request: ReviewRequest,
    ) -> str:
        """Replace variables in the Jenkins job name.

        This will replace the following variables:

        ``{branch_name}``:
            The unmodified branch name.

        ``{repository_name}``:
            The unmodified configured repository name.

        ``{noslash_branch_name}``:
            The branch name with slashes replaced with underscores.

        ``{noslash_repository_name}``:
            The configured repository name with slashes replaced with
            underscores.

        ``{branch}``:
            Legacy equivalent to ``{noslash_branch_name}``.

        ``{repository_name}``:
            Legacy equivalent to ``{noslash_repository_name}``.

        Version Changed:
            4.0:
            * Added support for ``{branch_name}``, ``{repository_name}``,
              ``{noslash_branch_name}``, and ``{noslash_repository_name}``.
            * The job name itself is no longer normalized, just specific
              variables within it.

        Args:
            job_name (str):
                The template string for the Jenkins job name.

            repository (reviewboard.scmtools.models.Repository):
                The repository for the change being built.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request for the change being built.

        Returns:
            str:
            The resulting job name.
        """
        branch = review_request.branch
        repo_name = repository.name
        noslash_branch = branch.replace('/', '_')
        noslash_repo_name = repo_name.replace('/', '_')

        var_map: dict[str, str] = {
            # Plain variables
            'branch_name': branch,
            'repository_name': repo_name,

            # No-slash variations
            'noslash_branch_name': noslash_branch,
            'noslash_repository_name': noslash_repo_name,

            # Legacy variables
            'branch': noslash_branch,
            'repository': noslash_repo_name,
        }

        return re.sub(
            r'\{(%s)\}' % '|'.join(var_map.keys()),
            lambda m: var_map[m.group(1)],
            job_name)



"""Base test cases for rbintegrations."""

from __future__ import annotations

from typing import Generic, TypeVar, cast

from kgb import SpyAgency
from reviewboard.extensions.testing import ExtensionTestCase
from reviewboard.integrations.base import (Integration,
                                           get_integration_manager)

from rbintegrations.extension import RBIntegrationsExtension


_IntegrationT = TypeVar('_IntegrationT', bound=Integration)


class RBIntegrationsExtensionTestCase(SpyAgency, ExtensionTestCase):
    """Base class for unit tests for rbintegrations."""

    extension_class = RBIntegrationsExtension


class IntegrationTestCase(RBIntegrationsExtensionTestCase,
                          Generic[_IntegrationT]):
    """Base class for unit tests for individual integrations."""

    #: The integration class to test.
    integration_cls: (type[_IntegrationT] | None) = None

    ######################
    # Instance variables #
    ######################

    #: The integration instance.
    integration: _IntegrationT

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        integration_mgr = get_integration_manager()

        integration_cls = self.integration_cls
        assert integration_cls is not None, (
            f'{type(self).__name__}.integration_cls must be set')

        assert integration_cls.integration_id is not None
        self.integration = cast(
            _IntegrationT,
            integration_mgr.get_integration(integration_cls.integration_id))

        integration_mgr.clear_all_configs_cache()



"""Integration for Matrix chat."""

from __future__ import annotations

import json
import logging
import re
from typing import MutableMapping, Optional, Sequence, TYPE_CHECKING
from urllib.request import Request, urlopen

from django.utils.translation import gettext_lazy as _
from django.utils.functional import cached_property
from django.utils.html import format_html
from reviewboard.admin.server import build_server_url

from rbintegrations.basechat.integration import BaseChatIntegration
from rbintegrations.matrix.forms import MatrixIntegrationConfigForm
from rbintegrations.util.compat.logs import log_timed

if TYPE_CHECKING:
    from djblets.util.typing import JSONDict
    from rbintegrations.basechat.integration import FieldsDict
    from reviewboard.reviews.models import ReviewRequest
    from reviewboard.site.models import LocalSite


logger = logging.getLogger(__name__)


class MatrixIntegration(BaseChatIntegration):
    """Integrates Review Board with Matrix.

    This will handle updating Matrix rooms when review requests are posted,
    changed, or closed, and when there's new activity on the review request.
    """

    name = 'Matrix'
    description = _(
        'Notifies rooms in Matrix when review requests are created, '
        'updated, and reviewed.'
    )

    default_settings = {
        'access_token': '',
        'notify_username': 'Review Board',
        'room': '',
    }

    config_form_cls = MatrixIntegrationConfigForm

    assets_base_url = (
        'https://static.reviewboard.org/integration-assets/matrix'
    )

    def initialize(self) -> None:
        """Initialize the integration."""
        super(MatrixIntegration, self).initialize()

        self._SHORTCODE_PATTERN = re.compile(r':[a-z_]+:')

    def notify(
        self,
        *,
        title: str,
        title_link: str,
        fallback_text: str,
        local_site: Optional[LocalSite],
        review_request: ReviewRequest,
        event_name: Optional[str] = None,
        fields: Sequence[FieldsDict] = [],
        pre_text: Optional[str] = None,
        body: Optional[str] = None,
        color: Optional[str] = None,
        thumb_url: Optional[str] = None,
        image_url: Optional[str] = None,
    ) -> None:
        """Send a notification to Matrix.

        This will post the given message to any Matrix rooms configured to
        receive it.

        Args:
            title (str):
                The title for the message.

            title_link (str):
                The link for the title of the message.

            fallback_text (str):
                The non-rich fallback text to display in the chat, for use in
                IRC and other services.

            local_site (reviewboard.site.models.LocalSite):
                The Local Site for the review request or review emitting
                the message. Only integration configurations matching this
                Local Site will be processed.

            review_request (reviewboard.reviews.models.ReviewRequest):
                The review request the notification is bound to.

            event_name (str, optional):
                The name of the event triggering this notification.

            fields (list of rbintegrations.basechat.integration.FieldsDict,
                    optional):
                The fields comprising the rich message to display in chat.

            pre_text (str, optional):
                Text to display before the rest of the message.

            body (str, optional):
                The body of the message.

            color (str, optional):
                A color string or RGB hex value for the message.

            thumb_url (str, optional):
                URL of an image to show on the side of the message.

            image_url (str, optional):
                URL of an image to show in the message.
        """
        message: list[str] = []
        message.append(
            format_html('<strong>{}</strong><p>{}</p>',
                        fallback_text, title))

        if not color:
            color = self.default_color

        if fields:
            for field in fields:
                text = self.replace_shortcode(field['value'])
                message.append(format_html(
                    '<strong><font color="{}">{}</font></strong><p>{}</p>',
                    color, field['title'], text))

        if body:
            message.append(format_html('<blockquote>{}</blockquote>', body))

        payload: JSONDict = {
            'body': '',
            'msgtype': 'm.text',
            'formatted_body': ''.join(message),
            'format': 'org.matrix.custom.html',
        }

        # Send a notification to any configured rooms.
        for config in self.get_configs(local_site):
            if not config.match_conditions(form_cls=self.config_form_cls,
                                           review_request=review_request):
                continue

            room_id = config.get('room_id')
            access_token = config.get('access_token')
            server = config.get('server')

            with log_timed(f'Sending notification for event "{event_name}", '
                           f'review_request ID {review_request.pk} to room '
                           f'"{room_id}"',
                           logger=logger) as log_timer:
                try:
                    data = json.dumps(payload).encode('utf-8')
                    url = (
                        f'{server}/_matrix/client/r0/rooms/{room_id}/send/'
                        f'm.room.message?access_token={access_token}'
                    )
                    headers: MutableMapping[str, str] = {
                        'Content-Length': str(len(data)),
                        'Content-Type': 'application/json',
                    }

                    urlopen(Request(url, data, headers))
                except Exception as e:
                    logger.exception('[%s] Failed to send notification: %s',
                                     log_timer.trace_id, e)

    def replace_shortcode(
        self,
        s: str,
    ) -> str:
        """Replaces supported shortcodes in a string with their unicodes.

        Args:
            s (str):
                The text that may contain shortcodes.

        Returns:
            str:
            The text with all supported shortcodes replaced by their
            equivalent Unicode characters.
        """
        emoji_unicode = {
            ':warning:': '\U000026A0',
            ':white_check_mark:': '\U00002705',
        }
        return self._SHORTCODE_PATTERN.sub(
            lambda x: emoji_unicode[x.group()], s)

    def format_link(
        self,
        *,
        path: str,
        text: str,
    ) -> str:
        """Format the given URL and text to be shown in a Matrix message.

        This will combine together the parts of the URL (method, domain, path)
        and format it using Matrix's URL syntax.

        Args:
            path (str):
                The path on the Review Board server.

            text (str):
                The text for the link.

        Returns:
            str:
            The link for use in Matrix.
        """
        text = text.replace('&', '&amp;')
        text = text.replace('<', '&lt;')
        text = text.replace('>', '&gt;')

        return '%s | %s' % (build_server_url(path), text)

    # NOTE: It's technically type-unsafe to replace an attribute with a
    #       property, but this works for us practically.
    @cached_property
    def icon_static_urls(self) -> dict[str, str]:  # type: ignore
        """Return the icons used for the integration.

        Returns:
            dict:
            The icons for Matrix.
        """
        from rbintegrations.extension import RBIntegrationsExtension

        extension = RBIntegrationsExtension.instance

        assert extension is not None

        return {
            '1x': extension.get_static_url('images/matrix/icon.png'),
            '2x': extension.get_static_url('images/matrix/icon@2x.png'),
        }



"""URLs for the GitLab CI integration.

Version Added:
    5.0
"""

from __future__ import annotations

from django.urls import path

from rbintegrations.gitlabci.views import GitlabCIWebHookView


urlpatterns = [
    path(
        'webhook/',
        GitlabCIWebHookView.as_view(),
        name='gitlab-ci-webhook',
    ),
]



"""Internal support for handling deprecations.

The version-specific objects in this module are not considered stable between
releases, and may be removed at any point. The base objects are considered
stable.

Version Added:
    4.0
"""

from housekeeping import BasePendingRemovalWarning, BaseRemovedInWarning


class PendingRemovalInRBIntegrationsWarning(BasePendingRemovalWarning):
    project = 'Review Board Integrations'


class BaseRemovedInRBIntegrationsWarning(BaseRemovedInWarning):
    project = 'Review Board Integrations'


class RemovedInRBIntegrations50Warning(BaseRemovedInRBIntegrationsWarning):
    version = '5.0'



"""Base form for configuring CI integrations.

Version Added:
    4.0
"""

from __future__ import annotations

from django import forms
from django.utils.translation import gettext_lazy as _
from djblets.forms.widgets import AmountSelectorWidget
from reviewboard.integrations.forms import IntegrationConfigForm

from rbintegrations.util.conditions import ReviewRequestConditionsField


class BaseCIIntegrationConfigForm(IntegrationConfigForm):
    """Base form for configuring CI integrations.

    Version Added:
        4.0
    """

    conditions = ReviewRequestConditionsField()

    run_manually = forms.BooleanField(
        label=_('Run builds manually'),
        required=False,
        help_text=_(
            'Wait to run this service until manually started. This will add '
            'a "Run" button to the build entry.'
        ),
        initial=False)

    timeout = forms.IntegerField(
        label=_('Build timeout'),
        required=False,
        help_text=_(
            'The amount of time until the build is considered to have timed '
            'out. If the build takes longer than this, it will be marked as '
            'timed out and can be re-run.'
        ),
        initial=None,
        widget=AmountSelectorWidget(
            unit_choices=[
                (1, _('seconds')),
                (60, _('minutes')),
                (60 * 60, _('hours')),
                (None, _('Never')),
            ],
            number_attrs={
                'min': 0,
            }))

    class Meta:
        pass



"""Unit tests for I Done This integration."""

import json
import logging
from io import StringIO
from urllib.error import HTTPError, URLError
from urllib.request import urlopen

from django.contrib.auth.models import User
from django.core.cache import cache
from django.test import RequestFactory
from djblets.cache.backend import cache_memoize, make_cache_key
from djblets.conditions import ConditionSet, Condition
from djblets.testing.decorators import add_fixtures
from reviewboard.reviews.conditions import ReviewRequestRepositoriesChoice
from reviewboard.reviews.models import ReviewRequestDraft
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)

from rbintegrations.idonethis.forms import (
    IDoneThisIntegrationAccountPageForm,
    IDoneThisIntegrationConfigForm)
from rbintegrations.idonethis.integration import IDoneThisIntegration, logger
from rbintegrations.idonethis.utils import get_user_team_ids
from rbintegrations.testing.testcases import IntegrationTestCase


class IDoneThisIntegrationTests(IntegrationTestCase):
    """Test posting of I Done This entries with review request activity."""

    integration_cls = IDoneThisIntegration

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self):
        """Set up this test case."""
        super(IDoneThisIntegrationTests, self).setUp()

        self.user = User.objects.create_user(username='testuser')
        self.team_ids_cache_key = make_cache_key('idonethis-team_ids-testuser')
        self.profile = self.user.get_profile()
        self.profile.settings['idonethis'] = {
            'api_token': encrypt_password('tok123'),
        }

    def test_post_review_request_closed_completed(self):
        """Testing IDoneThisIntegration posts on review request closed as
        completed
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        request = urlopen.spy.calls[0].args[0]
        self.assertEqual(request.get_full_url(),
                         'https://beta.idonethis.com/api/v2/entries')
        self.assertEqual(request.get_method(), 'POST')
        self.assertEqual(request.get_header('Authorization'), 'Token tok123')

        self.assertEqual(
            json.loads(request.data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/r/1/ #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_request_closed_completed_with_local_site(self):
        """Testing IDoneThisIntegration posts on review request closed as
        completed with local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)
        review_request.local_site.users.add(self.user)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)


        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/s/local-site-1/r/1/ #group',
            })

    def test_post_review_request_closed_discarded(self):
        """Testing IDoneThisIntegration posts on review request closed as
        discarded
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.DISCARDED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)


        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Discarded review request 1: Test Review Request '
                        'http://example.com/r/1/ #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_request_closed_discarded_with_local_site(self):
        """Testing IDoneThisIntegration posts on review request closed as
        discarded with local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)
        review_request.local_site.users.add(self.user)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.DISCARDED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)


        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Discarded review request 1: Test Review Request '
                        'http://example.com/s/local-site-1/r/1/ #group',
            })

    def test_post_review_request_published_normal(self):
        """Testing IDoneThisIntegration posts on review request published"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=False)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.publish(self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Published review request 1: Test Review Request '
                        'http://example.com/r/1/ #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_request_published_normal_with_local_site(self):
        """Testing IDoneThisIntegration posts on review request published with
        local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=False)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)
        review_request.local_site.users.add(self.user)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.publish(self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)


        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Published review request 1: Test Review Request '
                        'http://example.com/s/local-site-1/r/1/ #group',
            })

    def test_post_review_request_published_reopen(self):
        """Testing IDoneThisIntegration posts on review request published after
        discard and reopen
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        review_request.close(review_request.DISCARDED)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.reopen(self.user)

        # Reopened discarded review is not public until publish,
        # so post_entry should not post.
        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 0)

        review_request.publish(self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 2)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Reopened review request 1: Test Review Request '
                        'http://example.com/r/1/ #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_request_published_reopen_with_local_site(self):
        """Testing IDoneThisIntegration posts on review request published after
        discard and reopen with local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)
        review_request.local_site.users.add(self.user)

        review_request.close(review_request.DISCARDED)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.reopen(self.user)

        # Reopened discarded review is not public until publish,
        # so post_entry should not post.
        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 0)

        review_request.publish(self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 2)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Reopened review request 1: Test Review Request '
                        'http://example.com/s/local-site-1/r/1/ #group',
            })

    def test_post_review_request_published_update(self):
        """Testing IDoneThisIntegration posts on review request published after
        update
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        draft = ReviewRequestDraft.create(review_request)
        draft.summary = 'My new summary'
        draft.description = 'My new description'
        draft.save()

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.publish(self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Updated review request 1: My new summary '
                        'http://example.com/r/1/ #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_request_published_update_with_local_site(self):
        """Testing IDoneThisIntegration posts on review request published after
        update with local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)
        review_request.local_site.users.add(self.user)

        draft = ReviewRequestDraft.create(review_request)
        draft.summary = 'My new summary'
        draft.description = 'My new description'
        draft.save()

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.publish(self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Updated review request 1: My new summary '
                        'http://example.com/s/local-site-1/r/1/ #group',
            })

    def test_post_review_request_reopened(self):
        """Testing IDoneThisIntegration posts on review request reopened after
        after submit
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        review_request.close(review_request.SUBMITTED)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.reopen(self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Reopened review request 1: Test Review Request '
                        'http://example.com/r/1/ #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_request_reopened_with_local_site(self):
        """Testing IDoneThisIntegration posts on review request reopened after
        submit with local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)
        review_request.local_site.users.add(self.user)

        review_request.close(review_request.SUBMITTED)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.reopen(self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Reopened review request 1: Test Review Request '
                        'http://example.com/s/local-site-1/r/1/ #group',
            })

    def test_post_review_published_with_0_issues(self):
        """Testing IDoneThisIntegration posts on review published with no open
        issues
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        review = self.create_review(review_request, user=self.user)
        self.create_general_comment(review)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted review on review request 1: '
                        'Test Review Request '
                        'http://example.com/r/1/#review1 #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_published_with_0_issues_local_site(self):
        """Testing IDoneThisIntegration posts on review published with no open
        issues and local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)

        review = self.create_review(review_request, user=self.user)
        self.create_general_comment(review)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted review on review request 1: '
                        'Test Review Request '
                        'http://example.com/s/local-site-1/r/1/#review1 '
                        '#group',
            })

    def test_post_review_published_with_1_issue(self):
        """Testing IDoneThisIntegration posts on review published with 1 open
        issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        review = self.create_review(review_request, user=self.user)
        self.create_general_comment(review, issue_opened=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted review (1 issue) on review request 1: '
                        'Test Review Request '
                        'http://example.com/r/1/#review1 #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_published_with_1_issue_local_site(self):
        """Testing IDoneThisIntegration posts on review published with 1 open
        issue and local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)

        review = self.create_review(review_request, user=self.user)
        self.create_general_comment(review, issue_opened=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted review (1 issue) on review request 1: '
                        'Test Review Request '
                        'http://example.com/s/local-site-1/r/1/#review1 '
                        '#group',
            })

    def test_post_review_published_with_2_issues(self):
        """Testing IDoneThisIntegration posts on review published with > 1 open
        issues
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        review = self.create_review(review_request, user=self.user)
        self.create_general_comment(review, issue_opened=True)
        self.create_general_comment(review, issue_opened=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted review (2 issues) on review request 1: '
                        'Test Review Request '
                        'http://example.com/r/1/#review1 #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_published_with_2_issues_local_site(self):
        """Testing IDoneThisIntegration posts on review published with > 1 open
        issues and local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)

        review = self.create_review(review_request, user=self.user)
        self.create_general_comment(review, issue_opened=True)
        self.create_general_comment(review, issue_opened=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted review (2 issues) on review request 1: '
                        'Test Review Request '
                        'http://example.com/s/local-site-1/r/1/#review1 '
                        '#group',
            })

    def test_post_review_published_with_shipit_0_issues(self):
        """Testing IDoneThisIntegration posts on review published with Ship it!
        and no open issues
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True)
        self.create_general_comment(review)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted Ship it! on review request 1: '
                        'Test Review Request '
                        'http://example.com/r/1/#review1 #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_published_with_shipit_0_issues_local_site(self):
        """Testing IDoneThisIntegration posts on review published with Ship it!
        and no open issues and local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True)
        self.create_general_comment(review)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted Ship it! on review request 1: '
                        'Test Review Request '
                        'http://example.com/s/local-site-1/r/1/#review1 '
                        '#group',
            })

    def test_post_review_published_with_shipit_1_issue(self):
        """Testing IDoneThisIntegration posts on review published with Ship it!
        and 1 open issue
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True)
        self.create_general_comment(review, issue_opened=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted Ship it! (1 issue) on review request 1: '
                        'Test Review Request '
                        'http://example.com/r/1/#review1 #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_published_with_shipit_1_issue_local_site(self):
        """Testing IDoneThisIntegration posts on review published with Ship it!
        and 1 open issue and local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True)
        self.create_general_comment(review, issue_opened=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted Ship it! (1 issue) on review request 1: '
                        'Test Review Request '
                        'http://example.com/s/local-site-1/r/1/#review1 '
                        '#group',
            })

    def test_post_review_published_with_shipit_2_issues(self):
        """Testing IDoneThisIntegration posts on review published with Ship it!
        and > 1 open issues
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True)
        self.create_general_comment(review, issue_opened=True)
        self.create_general_comment(review, issue_opened=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted Ship it! (2 issues) on review request 1: '
                        'Test Review Request '
                        'http://example.com/r/1/#review1 #group',
            })

    @add_fixtures(['test_site'])
    def test_post_review_published_with_shipit_2_issues_local_site(self):
        """Testing IDoneThisIntegration posts on review published with Ship it!
        and > 1 open issues and local site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)

        review = self.create_review(review_request,
                                    user=self.user,
                                    ship_it=True)
        self.create_general_comment(review, issue_opened=True)
        self.create_general_comment(review, issue_opened=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Posted Ship it! (2 issues) on review request 1: '
                        'Test Review Request '
                        'http://example.com/s/local-site-1/r/1/#review1 '
                        '#group',
            })

    def test_post_reply_published(self):
        """Testing IDoneThisIntegration posts on reply published"""
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review, user=self.user)
        self.create_general_comment(reply, reply_to=comment)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        reply.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Replied to review request 1: Test Review Request '
                        'http://example.com/r/1/#review2 #group',
            })

    @add_fixtures(['test_site'])
    def test_post_reply_published_with_local_site(self):
        """Testing IDoneThisIntegration posts on reply published with local
        site
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            with_local_site=True,
            local_id=1,
            publish=True)
        group = self.create_review_group(name='group', with_local_site=True)
        review_request.target_groups.add(group)

        review = self.create_review(review_request,
                                    user=self.user,
                                    publish=True)
        comment = self.create_general_comment(review, issue_opened=True)

        reply = self.create_reply(review, user=self.user)
        self.create_general_comment(reply, reply_to=comment)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', with_local_site=True)
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        reply.publish()

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Replied to review request 1: Test Review Request '
                        'http://example.com/s/local-site-1/r/1/#review2 '
                        '#group',
            })

    def test_no_post_review_published_to_owner_only(self):
        """Testing IDoneThisIntegration doesn't post on review published to
        owner only
        """
        review_request = self.create_review_request(
            create_repository=True,
            summary='Test Review Request',
            publish=True)

        review = self.create_review(review_request, user=self.user)
        self.create_general_comment(review)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review.publish(to_owner_only=True)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 0)
        self.assertEqual(len(urlopen.spy.calls), 0)

    def test_no_post_without_user(self):
        """Testing IDoneThisIntegration doesn't post without a user"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(self.integration.get_configs)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(self.integration.get_configs.spy.calls), 0)
        self.assertEqual(len(urlopen.spy.calls), 0)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)

    def test_no_post_with_inactive_user(self):
        """Testing IDoneThisIntegration doesn't post with inactive user"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)

        self.user.is_active = False

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(self.integration.get_configs)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(self.integration.get_configs.spy.calls), 0)
        self.assertEqual(len(urlopen.spy.calls), 0)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)

    def test_no_post_without_api_token(self):
        """Testing IDoneThisIntegration doesn't post without an API Token"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)

        self.profile.settings['idonethis'] = {}

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(self.integration.get_configs)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(self.integration.get_configs.spy.calls), 0)
        self.assertEqual(len(urlopen.spy.calls), 0)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)

    def test_no_post_without_matching_condition(self):
        """Testing IDoneThisIntegration doesn't post without a matching
        condition
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123', repository_operator='none')
        self._create_config(team_id='teamABC', repository_operator='none')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(get_user_team_ids)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(get_user_team_ids.spy.calls), 0)
        self.assertEqual(len(urlopen.spy.calls), 0)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)

    def test_no_post_without_matching_team_id(self):
        """Testing IDoneThisIntegration doesn't post without a matching
        team ID
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123x')
        self._create_config(team_id='teamABCx')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(get_user_team_ids)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(get_user_team_ids.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 0)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)

    def test_no_post_with_team_ids_error(self):
        """Testing IDoneThisIntegration doesn't post after error while
        requesting the user team IDs
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)

        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.assertNotIn(self.team_ids_cache_key, cache)

        self.spy_on(self.integration.post_entry)
        self.spy_on(get_user_team_ids)
        self.spy_on(urlopen, call_fake=_urlopen_raise_httperror)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(get_user_team_ids.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)
        self.assertEqual(urlopen.spy.calls[0].args[0].get_full_url(),
                         'https://beta.idonethis.com/api/v2/teams')

    def test_try_post_with_httperror(self):
        """Testing IDoneThisIntegration tries to post to multiple teams with
        HTTPError
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self._create_config(team_id='teamABC')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_fake=_urlopen_raise_httperror)
        self.spy_on(logger.error)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertSpyCallCount(self.integration.post_entry, 1)
        self.assertSpyCallCount(urlopen, 2)
        self.assertSpyCallCount(logger.error, 2)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)

    def test_try_post_with_urlerror(self):
        """Testing IDoneThisIntegration tries to post to multiple teams with
        URLError
        """
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self._create_config(team_id='teamABC')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_fake=_urlopen_raise_urlerror)
        self.spy_on(logger.error)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertSpyCallCount(self.integration.post_entry, 1)
        self.assertSpyCallCount(urlopen, 2)
        self.assertSpyCallCount(logger.error, 2)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)

    def test_post_to_multiple_teams(self):
        """Testing IDoneThisIntegration posts to multiple matched teams"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self._create_config(team_id='teamABC')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(get_user_team_ids)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(get_user_team_ids.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 2)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/r/1/ #group',
            })

        self.assertEqual(
            json.loads(urlopen.spy.calls[1].args[0].data),
            {
                'team_id': 'teamABC',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/r/1/ #group',
            })

    def test_post_once_to_duplicate_teams(self):
        """Testing IDoneThisIntegration posts once to duplicate teams"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='group')
        review_request.target_groups.add(group)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self._create_config(team_id='teamABC')
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(get_user_team_ids)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(get_user_team_ids.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 2)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/r/1/ #group',
            })

        self.assertEqual(
            json.loads(urlopen.spy.calls[1].args[0].data),
            {
                'team_id': 'teamABC',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/r/1/ #group',
            })

    def test_post_with_multiple_target_groups(self):
        """Testing IDoneThisIntegration posts with multiple target groups"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)
        group = self.create_review_group(name='groupA')
        review_request.target_groups.add(group)
        group = self.create_review_group(name='new group-B')
        review_request.target_groups.add(group)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/r/1/ #groupA #new_group_B',
            })

    def test_post_without_target_groups(self):
        """Testing IDoneThisIntegration posts without target groups"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/r/1/',
            })

    def test_post_with_extra_whitespace_removed(self):
        """Testing IDoneThisIntegration posts with extra whitespace removed"""
        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='    Test    Review     Request    ',
            publish=True)

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})
        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.spy_on(self.integration.post_entry)
        self.spy_on(urlopen, call_original=False)

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)

        self.assertEqual(
            json.loads(urlopen.spy.calls[0].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/r/1/',
            })

    def test_post_multiple_events_with_single_team_ids_request(self):
        """Testing IDoneThisIntegration posts multiple events with a single
        request for the user team IDs
        """
        def _urlopen(request, **kwargs):
            if request.get_full_url().endswith('/teams'):
                return StringIO(json.dumps([
                    {
                        'hash_id': 'team123',
                    },
                ]))

            return StringIO('')

        review_request = self.create_review_request(
            create_repository=True,
            submitter=self.user,
            summary='Test Review Request',
            publish=False)

        self._create_config(team_id='team123')
        self.integration.enable_integration()

        self.assertNotIn(self.team_ids_cache_key, cache)

        self.spy_on(self.integration.post_entry)
        self.spy_on(get_user_team_ids)
        self.spy_on(urlopen, call_fake=_urlopen)

        review_request.publish(self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 1)
        self.assertEqual(len(get_user_team_ids.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 2)

        self.assertIsNone(self.integration.post_entry.spy.calls[0].exception)
        self.assertEqual(urlopen.spy.calls[0].args[0].get_full_url(),
                         'https://beta.idonethis.com/api/v2/teams')

        self.assertIn(self.team_ids_cache_key, cache)
        self.assertEqual(cache.get(self.team_ids_cache_key), {'team123'})

        self.assertEqual(urlopen.spy.calls[1].args[0].get_full_url(),
                         'https://beta.idonethis.com/api/v2/entries')

        self.assertEqual(
            json.loads(urlopen.spy.calls[1].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Published review request 1: Test Review Request '
                        'http://example.com/r/1/',
            })

        review_request.close(review_request.SUBMITTED, self.user)

        self.assertEqual(len(self.integration.post_entry.spy.calls), 2)
        self.assertEqual(len(get_user_team_ids.spy.calls), 2)
        self.assertEqual(len(urlopen.spy.calls), 3)

        self.assertEqual(urlopen.spy.calls[2].args[0].get_full_url(),
                         'https://beta.idonethis.com/api/v2/entries')
        self.assertEqual(
            json.loads(urlopen.spy.calls[2].args[0].data),
            {
                'team_id': 'team123',
                'status': 'done',
                'body': 'Completed review request 1: Test Review Request '
                        'http://example.com/r/1/',
            })

    def _create_config(self,
                       team_id,
                       with_local_site=False,
                       repository_operator='any'):
        """Create an integration config with the given parameters.

        Args:
            team_id (unicode):
                The team ID to post entries to.

            with_local_site (boolean, optional):
                Whether the configuration should be for a local site.

            repository_operator (unicode, optional):
                Operator for a repository choice to satisfy the condition.
        """
        choice = ReviewRequestRepositoriesChoice()

        condition_set = ConditionSet(conditions=[
            Condition(choice=choice,
                      operator=choice.get_operator(repository_operator)),
        ])

        if with_local_site:
            local_site = self.get_local_site(name=self.local_site_name)
        else:
            local_site = None

        config = self.integration.create_config(name='Config %s' % team_id,
                                                enabled=True,
                                                local_site=local_site)
        config.set('team_id', team_id)
        config.set('conditions', condition_set.serialize())
        config.save()


class IDoneThisIntegrationFormTests(IntegrationTestCase):
    """Test the admin configuration and user account page forms."""

    integration_cls = IDoneThisIntegration

    def setUp(self):
        """Initialize this test case."""
        super(IDoneThisIntegrationFormTests, self).setUp()

        self.request = RequestFactory().get('test')
        self.user = User.objects.create_user(username='testuser')
        self.team_ids_cache_key = make_cache_key('idonethis-team_ids-testuser')
        self.profile = self.user.get_profile()

    def test_admin_clean_team_id_whitespace(self):
        """Testing IDoneThisIntegration admin form, cleaning team ID strips
        whitespace
        """
        form = IDoneThisIntegrationConfigForm(
            integration=self.integration,
            request=self.request,
            data={
                'team_id': '  team123  ',
            })

        self.spy_on(form.clean_team_id)

        form.full_clean()

        self.assertEqual(len(form.clean_team_id.spy.calls), 1)
        self.assertEqual(form.cleaned_data['team_id'], 'team123')

    def test_admin_clean_team_id_validation(self):
        """Testing IDoneThisIntegration admin form, cleaning team ID containing
        slash raises validation error
        """
        form = IDoneThisIntegrationConfigForm(integration=self.integration,
                                              request=self.request)
        form.cleaned_data = {
            'team_id': 't/team123',
        }

        self.assertRaisesValidationError(
            'Team ID cannot contain slashes.',
            form.clean_team_id)

    def test_user_clean_token_empty(self):
        """Testing IDoneThisIntegration user form, cleaning empty API token
        skips validation
        """
        form = IDoneThisIntegrationAccountPageForm(
            page=None,
            request=self.request,
            user=self.user,
            data={
                'idonethis_api_token': '   ',
            })

        self.spy_on(urlopen)
        self.spy_on(form.clean_idonethis_api_token)

        form.full_clean()

        self.assertEqual(len(urlopen.spy.calls), 0)
        self.assertEqual(len(form.clean_idonethis_api_token.spy.calls), 1)

        self.assertEqual(form.cleaned_data['idonethis_api_token'], '')

    def test_user_clean_token_validation_request(self):
        """Testing IDoneThisIntegration user form, cleaning API token strips
        whitespace and performs validation API request
        """
        form = IDoneThisIntegrationAccountPageForm(
            page=None,
            request=self.request,
            user=self.user,
            data={
                'idonethis_api_token': '  tok123  ',
            })

        self.spy_on(urlopen, call_original=False)
        self.spy_on(logging.error)
        self.spy_on(form.clean_idonethis_api_token)

        form.full_clean()

        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(len(logging.error.spy.calls), 0)
        self.assertEqual(len(form.clean_idonethis_api_token.spy.calls), 1)

        request = urlopen.spy.calls[0].args[0]
        self.assertEqual(request.get_full_url(),
                         'https://beta.idonethis.com/api/v2/noop')
        self.assertEqual(request.get_method(), 'GET')
        self.assertEqual(request.get_header('Authorization'), 'Token tok123')
        self.assertIsNone(request.data)

        self.assertEqual(form.cleaned_data['idonethis_api_token'], 'tok123')

    def test_user_clean_token_validation_httperror(self):
        """Testing IDoneThisIntegration user form, cleaning API token with
        HTTPError raises validation error
        """
        form = IDoneThisIntegrationAccountPageForm(page=None,
                                                   request=self.request,
                                                   user=self.user)
        form.cleaned_data = {
            'idonethis_api_token': 'tok123',
        }

        self.spy_on(urlopen, call_fake=_urlopen_raise_httperror)
        self.spy_on(logging.error)

        self.assertRaisesValidationError(
            'Error validating the API Token. Make sure the token matches your '
            'I Done This Account Settings.',
            form.clean_idonethis_api_token)

        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(len(logging.error.spy.calls), 1)

    def test_user_clean_token_validation_urlerror(self):
        """Testing IDoneThisIntegration user form, cleaning API token with
        URLError raises validation error
        """
        form = IDoneThisIntegrationAccountPageForm(page=None,
                                                   request=self.request,
                                                   user=self.user)
        form.cleaned_data = {
            'idonethis_api_token': 'tok123',
        }

        self.spy_on(urlopen, call_fake=_urlopen_raise_urlerror)
        self.spy_on(logging.error)

        self.assertRaisesValidationError(
            'Error validating the API Token. Make sure the token matches your '
            'I Done This Account Settings.',
            form.clean_idonethis_api_token)

        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(len(logging.error.spy.calls), 1)

    def test_user_load_token_empty(self):
        """Testing IDoneThisIntegration user form, loading empty API token
        when not in settings
        """
        form = IDoneThisIntegrationAccountPageForm(page=None,
                                                   request=self.request,
                                                   user=self.user)

        self.spy_on(decrypt_password)

        form.load()

        self.assertEqual(len(decrypt_password.spy.calls), 0)

        self.assertIsNone(form.fields['idonethis_api_token'].initial)

    def test_user_load_token_decrypted(self):
        """Testing IDoneThisIntegration user form, loading decrypted API token
        from settings
        """
        form = IDoneThisIntegrationAccountPageForm(page=None,
                                                   request=self.request,
                                                   user=self.user)
        self.profile.settings['idonethis'] = {
            'api_token': encrypt_password('tok123'),
        }

        self.spy_on(decrypt_password)

        form.load()

        self.assertEqual(len(decrypt_password.spy.calls), 1)

        self.assertEqual(form.fields['idonethis_api_token'].initial, 'tok123')

    def test_user_save_token_empty(self):
        """Testing IDoneThisIntegration user form, saving empty API token
        creates empty settings
        """
        form = IDoneThisIntegrationAccountPageForm(page=None,
                                                   request=self.request,
                                                   user=self.user)
        form.cleaned_data = {
            'idonethis_api_token': '',
        }

        self.spy_on(encrypt_password)
        self.spy_on(self.profile.save)

        form.save()

        self.assertEqual(len(encrypt_password.spy.calls), 0)
        self.assertEqual(len(self.profile.save.spy.calls), 1)

        self.assertEqual(self.profile.settings['idonethis'], {})

    def test_user_save_token_empty_removes_old_token(self):
        """Testing IDoneThisIntegration user form, saving empty API token
        removes old token from settings and deletes cached team IDs
        """
        form = IDoneThisIntegrationAccountPageForm(page=None,
                                                   request=self.request,
                                                   user=self.user)
        form.cleaned_data = {
            'idonethis_api_token': '',
        }
        self.profile.settings['idonethis'] = {
            'api_token': encrypt_password('tok123'),
            'other_key': 'other value',
        }

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})

        self.spy_on(encrypt_password)
        self.spy_on(self.profile.save)

        form.save()

        self.assertEqual(len(encrypt_password.spy.calls), 0)
        self.assertEqual(len(self.profile.save.spy.calls), 1)

        self.assertEqual(
            self.profile.settings['idonethis'],
            {
                'other_key': 'other value',
            })
        self.assertNotIn(self.team_ids_cache_key, cache)

    def test_user_save_token_encrypted(self):
        """Testing IDoneThisIntegration user form, saving encrypted API token
        to settings
        """
        form = IDoneThisIntegrationAccountPageForm(page=None,
                                                   request=self.request,
                                                   user=self.user)
        form.cleaned_data = {
            'idonethis_api_token': 'tok123',
        }

        self.spy_on(encrypt_password)
        self.spy_on(self.profile.save)

        form.save()

        self.assertEqual(len(encrypt_password.spy.calls), 1)
        self.assertEqual(len(self.profile.save.spy.calls), 1)

        self.assertEqual(
            decrypt_password(self.profile.settings['idonethis']['api_token']),
            'tok123')

    def test_user_save_token_encrypted_replaces_old_token(self):
        """Testing IDoneThisIntegration user form, saving encrypted API token
        replaces old token in settings and deletes cached team IDs
        """
        form = IDoneThisIntegrationAccountPageForm(page=None,
                                                   request=self.request,
                                                   user=self.user)
        form.cleaned_data = {
            'idonethis_api_token': 'tokABC',
        }
        self.profile.settings['idonethis'] = {
            'api_token': encrypt_password('tok123'),
            'other_key': 'other value',
        }

        cache.set(self.team_ids_cache_key, {'team123', 'teamABC'})

        self.spy_on(encrypt_password)
        self.spy_on(self.profile.save)

        form.save()

        self.assertEqual(len(encrypt_password.spy.calls), 1)
        self.assertEqual(len(self.profile.save.spy.calls), 1)

        self.assertEqual(
            decrypt_password(self.profile.settings['idonethis']['api_token']),
            'tokABC')
        self.assertEqual(self.profile.settings['idonethis']['other_key'],
                         'other value')
        self.assertNotIn(self.team_ids_cache_key, cache)


class IDoneThisIntegrationUtilTests(IntegrationTestCase):
    """Test utility methods which are not fully covered by other tests."""

    integration_cls = IDoneThisIntegration

    def setUp(self):
        """Initialize this test case."""
        super(IDoneThisIntegrationUtilTests, self).setUp()

        self.user = User.objects.create_user(username='testuser')
        self.team_ids_cache_key = make_cache_key('idonethis-team_ids-testuser')
        self.profile = self.user.get_profile()
        self.profile.settings['idonethis'] = {
            'api_token': encrypt_password('tok123'),
        }

    def test_get_user_team_ids_without_token(self):
        """Testing IDoneThisIntegration team IDs get None without an API
        token
        """
        self.profile.settings['idonethis'] = {}

        self.spy_on(cache_memoize)
        self.spy_on(urlopen)

        team_ids = get_user_team_ids(self.user)

        self.assertEqual(len(cache_memoize.spy.calls), 0)
        self.assertEqual(len(urlopen.spy.calls), 0)

        self.assertIsNone(team_ids)
        self.assertNotIn(self.team_ids_cache_key, cache)

    def test_get_user_team_ids_from_cache(self):
        """Testing IDoneThisIntegration team IDs get data from cache without
        API request
        """
        cached_team_ids = {'team123', 'teamABC'}
        cache.set(self.team_ids_cache_key, cached_team_ids)

        self.spy_on(cache_memoize)
        self.spy_on(urlopen)

        team_ids = get_user_team_ids(self.user)

        self.assertEqual(len(cache_memoize.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 0)

        self.assertEqual(team_ids, cached_team_ids)

    def test_get_user_team_ids_request(self):
        """Testing IDoneThisIntegration team IDs get data from API request
        if not in cache
        """
        response = json.dumps([
            {
                'name': 'Example Team 1',
                'created_at': '2016-07-05T07:13:57.873Z',
                'updated_at': '2016-07-05T07:13:57.873Z',
                'hash_id': 'team123'
            }, {
                'name': 'Example Team 2',
                'created_at': '2016-07-05T07:13:57.875Z',
                'updated_at': '2016-07-05T07:13:57.875Z',
                'hash_id': 'teamABC'
            }, {
                'name': 'Duplicate Team 1',
                'created_at': '2016-07-05T07:13:57.877Z',
                'updated_at': '2016-07-05T07:13:57.877Z',
                'hash_id': 'team123'
            }])

        self.spy_on(cache_memoize)
        self.spy_on(urlopen,
                    call_fake=lambda request, **kwargs: StringIO(response))
        self.spy_on(logging.error)

        team_ids = get_user_team_ids(self.user)

        self.assertEqual(len(cache_memoize.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(len(logging.error.spy.calls), 0)

        request = urlopen.spy.calls[0].args[0]
        self.assertEqual(request.get_full_url(),
                         'https://beta.idonethis.com/api/v2/teams')
        self.assertEqual(request.get_method(), 'GET')
        self.assertEqual(request.get_header('Authorization'), 'Token tok123')
        self.assertIsNone(request.data)

        self.assertEqual(team_ids, {'team123', 'teamABC'})
        self.assertEqual(team_ids, cache.get(self.team_ids_cache_key))

    def test_get_user_team_ids_request_empty_team_list(self):
        """Testing IDoneThisIntegration team IDs get empty set if API request
        gets empty list
        """
        self.spy_on(cache_memoize)
        self.spy_on(urlopen,
                    call_fake=lambda request, **kwargs: StringIO('[]'))
        self.spy_on(logging.error)

        team_ids = get_user_team_ids(self.user)

        self.assertEqual(len(cache_memoize.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(len(logging.error.spy.calls), 0)

        self.assertEqual(team_ids, set())
        self.assertEqual(team_ids, cache.get(self.team_ids_cache_key))

    def test_get_user_team_ids_request_httperror(self):
        """Testing IDoneThisIntegration team IDs get None if API request gets
        HTTPError
        """
        self.spy_on(cache_memoize)
        self.spy_on(urlopen, call_fake=_urlopen_raise_httperror)
        self.spy_on(logging.error)

        team_ids = get_user_team_ids(self.user)

        self.assertEqual(len(cache_memoize.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(len(logging.error.spy.calls), 2)

        self.assertIsNone(team_ids)
        self.assertNotIn(self.team_ids_cache_key, cache)

    def test_get_user_team_ids_request_urlerror(self):
        """Testing IDoneThisIntegration team IDs get None if API request gets
        URLError
        """
        self.spy_on(cache_memoize)
        self.spy_on(urlopen, call_fake=_urlopen_raise_urlerror)
        self.spy_on(logging.error)

        team_ids = get_user_team_ids(self.user)

        self.assertEqual(len(cache_memoize.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(len(logging.error.spy.calls), 2)

        self.assertIsNone(team_ids)
        self.assertNotIn(self.team_ids_cache_key, cache)

    def test_get_user_team_ids_request_invalid_json(self):
        """Testing IDoneThisIntegration team IDs get None if API request gets
        invalid JSON
        """
        self.spy_on(cache_memoize)
        self.spy_on(urlopen,
                    call_fake=lambda request, **kwargs: StringIO('[invalid'))
        self.spy_on(logging.error)

        team_ids = get_user_team_ids(self.user)

        self.assertEqual(len(cache_memoize.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(len(logging.error.spy.calls), 1)

        self.assertIsNone(team_ids)
        self.assertNotIn(self.team_ids_cache_key, cache)

    def test_get_user_team_ids_request_invalid_team_data(self):
        """Testing IDoneThisIntegration team IDs get None if API request gets
        invalid team data
        """
        self.spy_on(cache_memoize)
        self.spy_on(urlopen,
                    call_fake=lambda request, **kw: StringIO('[{"a":"b"}]'))
        self.spy_on(logging.error)

        team_ids = get_user_team_ids(self.user)

        self.assertEqual(len(cache_memoize.spy.calls), 1)
        self.assertEqual(len(urlopen.spy.calls), 1)
        self.assertEqual(len(logging.error.spy.calls), 1)

        self.assertIsNone(team_ids)
        self.assertNotIn(self.team_ids_cache_key, cache)


def _urlopen_raise_httperror(request, **kwargs):
    """Fake urlopen that raises an HTTPError for testing.

    Args:
        request (urllib2.Request):
            The request to open.

        **kwargs (dict):
            Additional keyword arguments passed to urlopen.

    Raises:
        urllib2.HTTPError:
            The error for testing.
    """
    raise HTTPError(request.get_full_url(), 401, '', {}, StringIO(''))


def _urlopen_raise_urlerror(request, **kwargs):
    """Fake urlopen that raises a URLError for testing.

    Args:
        request (urllib2.Request):
            The request to open.

        **kwargs (dict):
            Additional keyword arguments passed to urlopen.

    Raises:
        urllib2.URLError:
            The error for testing.
    """
    raise URLError('url error')
